"""
Utilities for hashing string and files.
"""
import pathlib
import hashlib


def hash_items(*items, chunk_size=4096, hashfunc=hashlib.md5, sort=True):
    """Create a unique text string hash from the given item objects.

    Parameters
    ----------
    *items : Tuple[obj, str, bytes, :obj:`pathlib.Path`]
        Items to use in calculating the hash.
    chunk_size : Optional[int]
        When reading the contents of files (from :obj:`pathlib.Path` items),
        read the files in the given number of chunk bytes.
    hashfunc : Optional[func]
        The type of hash to use.
    sort : Optional[bool]
        If True, sort the items before calculating the hash. Enabling this
        option ensures that the items order does not change the hash.

    Returns
    -------
    hashdigest : str
        The hash digest string.
    """

    # Create the hash function
    hashes = list()

    for item in items:
        if isinstance(item, pathlib.Path):
            if item.suffix == '.pdf':
                hashtxt = hash_pdf(item, chunk_size=chunk_size,
                                   hashfunc=hashfunc)
            else:
                hashtxt = hash_file(item, chunk_size=chunk_size,
                                    hashfunc=hashfunc)
            hashes.append(hashtxt)
        elif isinstance(item, bytes):
            hashes.append(hashfunc(item).digest())
        else:
            hashes.append(hashfunc(str(item).encode()).digest())

    if sort:
        hashes = sorted(hashes)

    hashobj = hashfunc()
    for item in hashes:
        hashobj.update(item)
    return hashobj.hexdigest()


def hash_file(filepath, chunk_size, hashfunc):
    """Create a unique hash for the file contents of the given filepath.

    Parameters
    ----------
    filepath : :obj:`pathlib.Path`
        The filepath of the file whose contents will be hashed.
    chunk_size : Optional[int]
        When reading the contents of files (from :obj:`pathlib.Path` items),
        read the files in the given number of chunk bytes.
    hashfunc : Optional[func]
        The type of hash to use.

    Returns
    -------
    hash : bytes
        The hash bytes.
    """
    # Create the hash function
    hashobj = hashfunc()

    # open file for reading in binary mode
    with open(filepath, 'rb') as file:
        # loop till the end of the file
        chunk = 0
        while chunk != b'':
            chunk = file.read(chunk_size)
            hashobj.update(chunk)
    return hashobj.digest()


def line_splitter(file, newline, chunk_size, tail=None):
    """Given a file object, read its contents in chunks into lines without
    breaking newlines."""
    tail = tail if tail is not None else ''
    while True:
        chunk = file.read(chunk_size)
        if not chunk:
            if tail:
                yield tail
            break
        lines = (tail + chunk).split(newline)
        tail = lines.pop(0)
        if lines:
            yield tail
            tail = lines.pop()
            yield from lines


def hash_pdf(filepath, chunk_size, hashfunc,
             startswith=(b'/CreationDate', b'/ModDate', b'/ID')):
    """Create a unique hash for the file contents of the given pdf filepath.

    PDF files are parsed differently because they contain metadata on the
    date created, which may not reflect a change in the actual content of the
    file.

    Parameters
    ----------
    filepath : :obj:`pathlib.Path`
        The filepath of the file whose contents will be hashed.
    chunk_size : Optional[int]
        When reading the contents of files (from :obj:`pathlib.Path` items),
        read the files in the given number of chunk bytes.
    hashfunc : Optional[func]
        The type of hash to use.
    startswith : Optional[Tuple[str]]
        Lines that start with the given bytes will be ignored in the hash.

    Returns
    -------
    hash : bytes
        The hash bytes.
    """
    # Create the hash function
    hashobj = hashfunc()

    # Retrieve the lines from the file
    with open(filepath, 'rb') as file:
        for line in line_splitter(file, newline=b'\n', chunk_size=chunk_size,
                                  tail=b''):
            if any(line.startswith(starter) for starter in startswith):
                continue
            hashobj.update(line)

    return hashobj.digest()



"""
Utilities for formatting tex strings and text.
"""
from .exceptions import FormattingError
from ..attributes import Attributes
from ..utils.string import space_indent
from .. import settings


class TexFormatError(FormattingError):
    """Error in latex formatting."""
    pass


def tex_cmd(cmd, attributes='', formatted_content=None, indent=None):
    """Format a tex command.

    Parameters
    ----------
    cmd : Optional[str]
        The name of the LaTeX command to format.
    attributes : Optional[Union[:obj:`Attributes <.Attributes>`, str]]
        The attributes of the tag.
    formatted_content : Optional[str]
        The contents of the tex environment formatted as a string in LaTeX.
        If not specified, the tex_str will not be used as a LaTeX parameter
    indent : Optional[int]
        If specified, indent lines by the given number of spaces.

    Returns
    -------
    tex_env : str
        The LaTeX environment string

    Raises
    ------
    TexFormatError : :exc:`TexFormatError`
        A TexFormatError is raised if an non-allowed environment is used.
    """
    # Make sure the environment is permitted
    if (cmd not in settings.tex_cmd_arguments and
       cmd not in settings.tex_cmd_optionals):
        msg = "Cannot use the LaTeX command '{}'"
        raise TexFormatError(msg.format(cmd))

    # Format the attributes
    attributes = (Attributes(attributes) if isinstance(attributes, str) else
                  attributes)

    # Get the required arguments
    if cmd in settings.tex_cmd_arguments:
        reqs = attributes.filter(attrs=settings.tex_cmd_arguments[cmd],
                                 sort_by_attrs=True, target='tex')
        reqs_str = reqs.tex_arguments
    else:
        reqs = None
        reqs_str = ''

    # Make sure the correct number of required arguments were found
    if (reqs is not None and
       len(reqs) != len(settings.tex_cmd_arguments[cmd])):
        msg = ("The LaTeX environment '{}' did not receive the correct "
               "required arguments. Required arguments received: {}")
        raise TexFormatError(msg.format(cmd, reqs))

    # Get optional arguments
    if cmd in settings.tex_cmd_optionals:
        opts = attributes.filter(attrs=settings.tex_cmd_optionals[cmd],
                                 target='tex')
        opts_str = opts.tex_optionals
    else:
        opts_str = ''

    # Wrap the formatted_content in curly braces, if a formatted_content was
    # specified
    if isinstance(formatted_content, str):
        formatted_content = ("{" + formatted_content + "}"
                             if formatted_content else "{}")
    else:
        formatted_content = ''

    # format the tex command
    tex_text = "\\" + cmd + reqs_str + opts_str + formatted_content

    # Indent the text block, if specified
    if indent is not None:
        tex_text = space_indent(tex_text, number=indent)

    return tex_text


def tex_env(env, attributes, formatted_content, min_newlines=False,
            indent=None):
    """Format a tex environment.

    Parameters
    ----------
    env : str
        The name of the LaTeX environment to format.
    attributes : Union[:obj:`Attributes <.Attributes>`, str]
        The attributes of the tag.
    formatted_content : str
        The contents of the tex environment formatted as a string in LaTeX.
    min_newlines : Optional[bool]
        If True, extra new lines before, after and in the environment will not
        be included.
    indent : Optional[int]
        If specified, indent lines by the given number of spaces.

    Returns
    -------
    tex_env : str
        The LaTeX environment string

    Raises
    ------
    TexFormatError : :exc:`TexFormatError`
        A TexFormatError is raised if an non-allowed environment is used.
    """
    # Make sure the environment is permitted
    if (env not in settings.tex_env_arguments and
       env not in settings.tex_env_optionals):
        msg = "Cannot use the LaTeX environment '{}'"
        raise TexFormatError(msg.format(env))

    # Format the attributes
    attributes = (Attributes(attributes) if isinstance(attributes, str) else
                  attributes)

    # Get the required arguments
    # Get the required arguments
    if env in settings.tex_env_arguments:
        reqs = attributes.filter(attrs=settings.tex_env_arguments[env],
                                 target='tex',
                                 sort_by_attrs=True)
        reqs_str = reqs.tex_arguments
    else:
        reqs = None
        reqs_str = ''

    # Make sure the correct number of required arguments were found
    if (reqs is not None and
       len(reqs) != len(settings.tex_env_arguments[env])):
        msg = ("The LaTeX environment '{}' did not receive the correct "
               "required arguments. Required arguments received: {}")
        raise TexFormatError(msg.format(env, reqs))

    # Get optional arguments
    if env in settings.tex_env_optionals:
        opts = attributes.filter(attrs=settings.tex_env_optionals[env],
                                 target='tex',
                                 sort_by_attrs=True)
        opts_str = opts.tex_optionals
    else:
        opts_str = ''

    # Add a leading and trailing new line to the formatted_content,
    # if there isn't one
    formatted_content = ('\n' + formatted_content
                         if not formatted_content.startswith('\n') else
                         formatted_content)
    formatted_content = (formatted_content + '\n'
                         if not formatted_content.endswith('\n') else
                         formatted_content)

    # format the tex environment
    tex_text = ("\\begin" + '{' + env + '}' +
                reqs_str +
                opts_str +
                (' %' if min_newlines else '') +
                formatted_content +
                "\\end{" + env + "}")

    # Indent the text block, if specified
    if indent is not None:
        tex_text = space_indent(tex_text, number=indent)

    # Add new_lines on the two ends of the block, if not min_newlines
    if not min_newlines:
        tex_text = "\n" + tex_text + "\n"

    return tex_text


def tex_verb(formatted_content):
    """Format a tex verb command

    Parameters
    ----------
    formatted_content : str
        The contents of the tex environment formatted as a string in LaTeX.

    Returns
    -------
    tex_verb : str
        The LaTeX verb string
    """
    assert isinstance(formatted_content, str)

    if '|' not in formatted_content:
        return "\\verb|" + formatted_content + "|"
    elif '!' not in formatted_content:
        return "\\verb!" + formatted_content + "!"
    elif '^' not in formatted_content:
        return "\\verb^" + formatted_content + "^"
    return formatted_content



"""
Formatting of Table of Contents for documents
"""
from .headings import toc_levels as heading_toc_levels, Heading
from .ref import Ref
from .tag import Tag
from . import exceptions
from ..formats import xhtml_tag, xhtml_list


class TocError(Exception):
    """An error was encountered while processing a table of contents tag."""
    pass


class TocRef(Ref):
    """A Ref tag for the TOC.

    This is a separate class so that the label_fmt may be different for TOC
    entries.
    """
    # The tocref tag should not be accessible on its own; it is created
    # by the Toc tag
    active = False

    html_name = "li"

    def tex_fmt(self, content=None, attributes=None, mathmode=False,
                cache=None, level=1, **kwargs):
        list_level = self.attributes['level']
        tex_content = super().tex_fmt(content=content, attributes=attributes,
                                      mathmode=mathmode, cache=cache,
                                      level=level, **kwargs)
        return "§" * list_level + " " + tex_content + "\n"

    def html_fmt(self, content=None, attributes=None, cache=None,
                 format_func='html_fmt', method='html', level=1, **kwargs):
        # Wrap the tocref item in a list item
        html = super().html_fmt(content=content, attributes=attributes,
                                cache=cache, format_func=format_func,
                                method=method, level=level)

        tag_class = ('class="toc-level-{}"'.format(self.attributes['level'])
                     if 'level' in self.attributes else '')
        return xhtml_tag('li', formatted_content=html, attributes=tag_class,
                         method=method)


class Toc(Tag):
    """Table of contents and listings.

    contents : str
        The contents are the label types to list. The following entries
        are supported:

        - document: list the document labels
        - heading : list heading labels
        - figure : list figure labels
        - table : list table labels

        Additionally, the following modifiers have special meaning:
        - all : list labels from all documents
        - current : list labels from the current document (default)


        - expanded : pertains to 'all document' and 'all headings'.
                     Show all label references for all documents.
        - abbreviated : pertains to 'all document' and 'all headings'.
                        show all label references for the current document and
                        the top level heading for other documents.
        - collapsed : pertains to 'all document' and 'all headings'.
                     show only the documents without headings. (default)
    """

    active = True
    toc_kind = None
    toc_elements = None
    header_tag = None

    tex_env = "easylist"
    list_style = 'booktoc'
    html_name = 'ol'

    process_typography = False

    _ref_tags = None

    def __init__(self, name, content, attributes, context):
        super(Toc, self).__init__(name, content, attributes, context)

        # Get the TOC's kind from the tag's content
        content = self.content
        if isinstance(content, list) and len(content) > 0:
            content = content[0]
        if isinstance(content, str):
            self.toc_kind = content.split()
        elif isinstance(content, Tag):
            self.toc_kind = content.txt

        # If the TOC kind hasn't been assigned, the content could not be
        # parsed. Raise an exception
        if self.toc_kind is None:
            msg = "The {} tag could not parse the tag contents: {}"
            raise exceptions.TagError(msg.format(self.name, content))

        # Setup the TOC header, if specified
        header = self.attributes.pop('header', None)

        if header is not None:
            self.header_tag = Heading(name='TOC', content='Table of Contents',
                                      attributes='nolabel', context=context)

    def get_labels(self):
        """Get the labels, ordering function and labeling type.

        Returns
        -------
        labels : List[:obj:`.label_manager.types.Label`]
            The labels referenced by this TOC.
        """
        assert self.context.is_valid('label_manager')

        # If 'all' is specified in the toc_kind, then all documents should be
        # selected. This is done by having a doc_id of None with the
        # 'get_labels_by_kind' method of the label manager. If 'all' is not
        # specified, then use this document's doc_id. This will return labels
        # only for this document and its context from the 'get_labels_by_kind'
        # method of the label manager.
        context = self.context
        doc_id = context.get('doc_id') if 'all' not in self.toc_kind else None
        label_manager = self.context['label_manager']

        labels = []
        if 'heading' in self.toc_kind or 'headings' in self.toc_kind:
            labels += label_manager.get_labels_by_kind(doc_id=doc_id,
                                                       kinds='heading')
        if 'document' in self.toc_kind or 'documents' in self.toc_kind:
            labels += label_manager.get_labels_by_kind(doc_id=doc_id,
                                                       kinds='document')

        # Now filter apply additional filters
        doc_id = context.get('doc_id')
        if 'abbreviated' in self.toc_kind:
            current_doc_id = None
            filtered_labels = []

            for label in labels:
                if label.doc_id == doc_id:
                    # This is a label for the current document keep it
                    filtered_labels.append(label)
                elif label.doc_id != current_doc_id:
                    # This is a label for a different document. Only keep it
                    # if it's the first label for this other document
                    filtered_labels.append(label)
                current_doc_id = label.doc_id

            # Transfer the filtered list
            labels.clear()
            labels += filtered_labels

        return labels

    @property
    def reference_tags(self):
        """This tag's TocRef tag items.
        """
        labels = self.get_labels()

        # Collect the created tags.
        tags = []

        # Got through the labels and keep track of the levels
        current_level = 1
        for label in labels:
            # Get the level for the label
            if label.kind[-1] in heading_toc_levels:
                level = heading_toc_levels.index(label.kind[-1])
            else:
                level = current_level
            current_level = level

            # Create the tag and add it to the tags list
            tag_name = 'toc-' + label.kind[-1]
            tag = TocRef(name=tag_name, content=label.id,
                         attributes=self.attributes, context=self.context)
            tag.attributes['level'] = level

            tags.append(tag)

        return tags

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        tags = list(self.reference_tags)
        tags[0:0] = "\\ListProperties(Hide=2)\n"  # Add to front
        return super().tex_fmt(content=tags, attributes=self.list_style)

    def html_fmt(self, content=None, attributes=None, cache=None,
                 format_func='html_fmt', method='html', level=1, **kwargs):
        tags = self.reference_tags
        elements = []

        root_document = self.context.root_document
        documents_by_id = (root_document.documents_by_id(recursive=True)
                           if root_document is not None else None)

        # Get the labels
        labels = self.get_labels()
        labels_by_id = {label.id: label for label in labels}

        for tag in tags:
            listlevel = tag.attributes['level']
            label = labels_by_id.get(tag.label_id, None)

            cache = dict() if cache is None else cache
            cache['label'] = label
            cache['documents_by_id'] = documents_by_id

            func = getattr(tag, format_func)
            tag_html = func(cache=cache, method=method, level=level + 1)
            elements.append((listlevel, tag_html))

        return xhtml_list(*elements, attributes='class="toc"',
                          listtype=self.html_name, method=method, level=level)



"""Exceptions for Tag objects"""


class TagError(Exception):
    """An error was encountered while interpreting a tag."""
    pass


def assert_content_str(content):
    """Assert that the tag contents are a string, and raise an exception if
    it isn't.

    Raises
    ------
    TagError
        If the content isn't a string, raise a TagError.
    """
    if not isinstance(content, str):
        msg = ("The tag contents must be a string. "
               "The followed content was inserted: {}")
        raise TagError(msg.format(content))



"""
A builder that renders a string to a file.
"""
import pathlib
import logging

import jinja2
import jinja2.meta

from .builder import Builder
from .exceptions import BuildError
from .utils import generate_mock_parameters, generate_outfilepath
from ..paths import SourcePath
from ..paths.utils import find_file
from ..utils.list import uniq
from ..utils.classes import weakattr
from .. import settings


def run(template, context, outfilepath, target):
    """Run the command with the given arguments."""
    # rendered_string = asyncio.run(template.render_async(**context,
    #                               outfilepath=outfilepath, target=target))
    if 'target' in context:
        rendered_string = template.render(**context, outfilepath=outfilepath)
    else:
        rendered_string = template.render(**context, outfilepath=outfilepath,
                                          target=target)
    outfilepath.write_text(rendered_string)
    return outfilepath


@jinja2.pass_context
def rewrite_path(context, stub):
    """A Jinja2 filter for rewriting paths, like css paths.

    .. note:: This function will only modify the path if an existing file can
              be found in the target directory. If not, it will log an error
              and return the path unchanged. This means that the dependent
              .css (and other) files should have been copied to the target
              directory before rendering the template.

    .. note:: This function is not sped up from using a ThreadPoolExecutor,
              and implementing with a ProcessPoolExecutor gives problems
              with pickling objects.
    """
    # Strip leading slashes so that the stub is not an absolute path
    stub = stub.strip('/')

    # Find the file with respect to the outfilepath
    outfilepath = context['outfilepath']
    paths = [outfilepath.parent,  # the outfilepath directory
             outfilepath.use_subpath('')  # the target_root / target
             ]

    try:
        filepath = find_file(stub, context={'paths': paths})
    except FileNotFoundError:
        msg = "Could not find '{}' in the target directory"
        logging.error(msg.format(stub))
        return stub

    # Make it a relative path, if specified
    target = context['target']
    string = filepath.get_url(context=context, target=target)

    return str(string)


class JinjaRender(Builder):
    """A builder that renders a file using Jinja2.

    .. note:: The JinjaRender should not render the template until the build
              step to make sure the document and context are properly loaded
              first.

    Parameters
    ----------
    env: :obj:`.builders.Environment`
        The build environment
    parameters, args : Tuple[:obj:`.paths.SourcePath`, str, tuple, list]
        The input parameters (dependencies), including filepaths, for the build
    outfilepath : Optional[:obj:`.paths.TargetPath`]
        If specified, the path for the output file.
    context : dict
        A context to use with the renderer.
    render_ext : str
        The extension for the rendered file. Either this or the outfilepath
        must be specified. ex: '.tex'
    """

    available = True
    action = 'render'
    priority = 1000
    active_requirements = ('priority',)
    scan_parameters_on_init = False  # Scan after all parameters are loaded

    context = weakattr()

    infilepath_ext = '.render'  # dummy extension for find_builder_cls

    render_ext = None

    def __init__(self, env, context, render_ext=None, **kwargs):
        super().__init__(env, **kwargs)

        # Checks
        assert render_ext or self.outfilepath, ("Either a render_ext or an "
                                                "outfilepath must be "
                                                "specified")
        self.render_ext = render_ext or self.outfilepath.suffix
        self.context = context

    def jinja_environment(self):
        """The jinja environment."""
        if getattr(self.env, '_jinja_environment', None) is None:
            # Create the loaders
            dl = jinja2.PackageLoader('disseminate', 'templates')

            # Create the environment
            ae = jinja2.select_autoescape(['html', 'htm', 'xml'])
            env = jinja2.Environment(autoescape=ae, loader=dl,
                                     keep_trailing_newline=True,)
            env.filters['rewrite_path'] = rewrite_path

            self.env._jinja_environment = env
        return self.env._jinja_environment

    def template(self):
        """Retrieve the template from the context"""
        context = self.context
        # Get the template filepath and convert to a pathlib.Path
        template_filepath = context.get('template', 'default')
        template_filepath = pathlib.Path(template_filepath)

        # get the target extension without a period
        target = self.render_ext.strip('.')

        # create the list of template paths to search
        # path1. ex: 'default '/ 'html' / 'template' '.html'
        path1 = template_filepath / target / 'template'
        path1 = path1.with_suffix('.' + target)

        # path2. ex: 'default/xhtml/toc' / '.xhtml
        path2 = template_filepath.with_suffix('.' + target)

        # path3: ex: 'default/xhtml/toc.xhtml'
        path3 = template_filepath

        templates = list(map(str, [path1, path2, path3]))
        # Get the Jinja2 environment
        jinja_env = self.jinja_environment()

        # Retrieve the template
        template = jinja_env.get_or_select_template(templates)
        return template

    @property
    def parameters(self):
        parameters = list(self._parameters) if self._parameters else []
        context = self.context

        # Render the string and add it (as well as dependent template files)
        # to the list of returned input parameters
        if context is not None:
            # Retrieve the template
            jinja_env = self.jinja_environment()
            template = self.template()

            # Add the template filepaths to the infilepath dependencies.
            # These are added to the parameters so that if they are changed,
            # the decider can trigger a new build.
            filepaths = template_filepaths(template=template,
                                           environment=jinja_env)
            parameters += filepaths

            # Add the context filepaths to the infilepath dependencies
            filepaths = context_filepaths(filepaths)
            parameters += filepaths

            # Scan for additional dependencies
            parameters += self.env.scanner.scan(parameters=parameters)

            # add hashes (from tags) to context values
            parameters += sorted(v.hash for v in context.values()
                                 if hasattr(v, 'hash') and v.hash is not None)

        return uniq(parameters)

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    @property
    def outfilepath(self):
        outfilepath = self._outfilepath

        if outfilepath is None:
            parameters = self.parameters

            if parameters:
                # Create an a mock infilepath with hash from the parameters
                sourcepath = generate_mock_parameters(env=self.env,
                                                      context=self.context,
                                                      parameters=parameters,
                                                      ext='.render')

                # Generate an outfilepath from the mock parameters
                outfilepath = generate_outfilepath(env=self.env,
                                                   parameters=sourcepath,
                                                   append=self.outfilepath_ext,
                                                   ext=self.render_ext,
                                                   target=self.target,
                                                   use_cache=self.use_cache,
                                                   use_media=self.use_media)

        # Make sure the outfilepath directory exists
        if outfilepath and not outfilepath.parent.is_dir():
            outfilepath.parent.mkdir(parents=True, exist_ok=True)

        return outfilepath

    @outfilepath.setter
    def outfilepath(self, value):
        self._outfilepath = value

    # Note: Enable this function and comment out the build to run the
    # render in the thread pool. This slows the run, however, by about 30-40%
    # def run_cmd(self, *args):
    #     if self.future is None:
    #         template = self.template()
    #         outfilepath = self.outfilepath
    #
    #         logging.debug("Rendering '{}' with Jinja2 "
    #                       "'{}'".format(outfilepath, template))
    #         future = executor.submit(run, template=template,
    #                                  context=self.context,
    #                                  outfilepath=outfilepath,
    #                                  target=self.render_ext)
    #         self.future = future

    def build(self, complete=False):
        template = self.template()
        context = self.context
        outfilepath = self.outfilepath

        logging.debug("Rendering '{}' with Jinja2 "
                      "'{}'".format(outfilepath, template))
        if 'target' in context:
            rendered_string = template.render(**context,
                                              outfilepath=outfilepath)
        else:
            rendered_string = template.render(**context,
                                              target=self.render_ext,
                                              outfilepath=outfilepath)

        outfilepath.write_text(rendered_string)
        self.build_needed(reset=True)
        return self.status

    @staticmethod
    def runtime_success(future):
        """Test whether a future from a subprocess is successful."""
        return future.result().exists()  # outfilepath exists

    @staticmethod
    def runtime_error(future, error_msg=None, raise_error=True):
        """Raise an error from a future working with subprocess"""
        outfilepath = future.result()

        if error_msg is None:
            error_msg = ("The render of  '{}' was "
                         "unsuccessful.".format(outfilepath))
        if raise_error:
            e = BuildError(error_msg)
            raise e
        else:
            logging.warning(error_msg)


# Utilities

def template_filepaths(template, environment):
    """Return a list of filepaths from a Jinja2 template object.

    Parameters
    ----------
    template : :obj:`jinja2.Template`
        The jinja2 template object to load filepaths for
    environment : :obj:`jinja2.Environment`
        The jinja2 environment object

    Returns
    -------
    filepaths : List[:obj:`pathlib.Path`]
        A list of paths.
    """
    # Prepare a list of absolute paths (render paths) for the templates
    filenames = list()

    # Get the loader from the environment
    loader = environment.loader

    # Find the parent templates for the template
    # Get the template name
    name = template.name

    # Get the template's filename and add it to the filenames set
    template_filename = pathlib.Path(template.filename)
    filenames.append(SourcePath(project_root=template_filename.parent,
                                subpath=template_filename.name))

    # Load the source code for the template using the loader
    source = loader.get_source(environment, name)

    # Produce a Jinja2 AST from the source
    ast = environment.parse(source)

    # Get a list of all parent template names from the AST
    parent_names = list(jinja2.meta.find_referenced_templates(ast))

    # Convert the parent names to template file paths (render paths)
    # This is done by loading the parent template objects.
    parent_templates = [environment.get_template(parent_name)
                        for parent_name in parent_names]

    # Run this function on those templates
    for parent_template in parent_templates:
        filenames += template_filepaths(parent_template, environment)

    return filenames


def context_filepaths(template_filepaths):
    """Return a list of context modifier files (ex: context.txt) from template
    filepaths.

    Parameters
    ----------
    template_filepaths : List[:obj:`pathlib.Path`]
        A list of filepaths for templates.

    Returns
    -------
    filepaths : List[:obj:`pathlib.Path`]
        A list of paths.
    """
    filepaths = []
    # Construct the context filepaths from the template_filepaths
    for filepath in template_filepaths:
        # Construct a test context filename1
        # ex: filepath: 'templates/default/html'
        #     context_fp: templates/default/context.txt
        context_fp = SourcePath(project_root=filepath.parent.parent,
                                subpath=settings.template_context_filename)

        # See if the test context filename exists and add it to the list
        # of context_filepaths if it isn't in there already.
        if (context_fp.is_file() and
                context_fp not in filepaths):
            filepaths.append(context_fp)
    return filepaths



"""
A builder to render a tex file to pdf.
"""
from .composite_builders import SequentialBuilder


class PdfRender(SequentialBuilder):
    """Render a tex file and render the pdf."""

    available = True
    priority = 1000

    infilepath_ext = '.render'  # dummy extension for find_builder_cls
    outfilepath_ext = '.pdf'

    def __init__(self, env, context=None, template=None, parameters=None,
                 outfilepath=None, subbuilders=None, use_cache=None,
                 **kwargs):

        # Setup the arguments
        parameters = parameters or []
        subbuilders = subbuilders or []

        # Setup a render if a context is specified
        if context is not None:
            render_cls = self.find_builder_cls(in_ext='.render')
            render_build = render_cls(env, context=context, render_ext='.tex',
                                      template=template, parameters=parameters,
                                      use_cache=True, **kwargs)

            # Set the parameters for this builder to match the render_build, if
            # used, so that the Md5Decision is properly calculated
            parameters += render_build.parameters

            subbuilders.append(render_build)
            self.subbuilder_for_outfilename = render_build

        # Setup a pdf builder
        pdf_build_cls = self.find_builder_cls(in_ext='.tex', out_ext='.pdf')
        pdf_build = pdf_build_cls(env, use_cache=True, **kwargs)
        subbuilders.append(pdf_build)
        super().__init__(env, parameters=parameters, outfilepath=outfilepath,
                         subbuilders=subbuilders, use_cache=use_cache,
                         **kwargs)



"""
The receivers for tag events.
"""
from .hash import process_hash
from .macros import process_macros
from .content import process_content
from .typography import process_typography
from .paragraphs import process_paragraphs

__all__ = ('process_hash', 'process_macros', 'process_content',
           'process_typography', 'process_paragraphs')
